"""file: ex1.py -- Correction TD2 -- bubble sort

Implements bubble sort.

"""

import random

"""bubble_sort_naive(li) returns a sorted version of 'li' by the
method of bubbling.

It is naive because it does not detect if the list is already sorted
at some point.

"""
def bubble_sort_naive_in_place(li):
  """Careful! This function modifies effectively the list li.

  """
  for i in range(len(li)):
    for j in range(len(li)-1):
      if li[j] > li[j+1]: li[j],li[j+1] = li[j+1],li[j]
  return li

def bubble_sort_naive(li):
  """This function sorts a copy of li and returns it. li stays
  unchanged.

  """
  tmp = li[:] # copy li
  for i in range(len(li)):
    for j in range(len(li)-1):
      if tmp[j] > tmp[j+1]: tmp[j],tmp[j+1] = tmp[j+1],tmp[j]
  return tmp

""" Same as before but test if some swap happened; if not, the list is
sorted, we can stop.

"""
def bubble_sort_better(li):
  tmp = li[:] # copy li
  for i in range(len(li)):
    swap = False
    for j in range(len(li)-1):
      if tmp[j] > tmp[j+1]:
        tmp[j],tmp[j+1] = tmp[j+1],tmp[j]
        swap = True
    if not swap: break
  return tmp

def bubble_sort_best(li):
  tmp,n = li[:],len(li) # copy li
  pos = n-1
  for i in range(n):
    swap = False 
    for j in range(pos):
      if tmp[j] > tmp[j+1]:
        tmp[j],tmp[j+1] = tmp[j+1],tmp[j]
        swap,pos = True, j
    if not swap: break
  return tmp

# another way, making an auxiliary function

"""bubble_up(li,swap) makes the maximum of li[:swap] go to the cell
indexed by swap-1. It returns the new last swap position after it.

"""
def bubble_up(li,swap):
  new = swap-1 # this is the maximum it can be
  for j in range(swap-1):
    if li[j] > li[j+1]:
      li[j],li[j+1] = li[j+1],li[j]
      new = j+1
  return new

"""bubble_sort(li) returns a sorted version of 'li' by the method of
bubbling.

"""
def bubble_sort(li):
  tmp = li[:] # copy li
  lswap = n # remember the last swap
  while lswap > 1:
    lswap = bubble_up(tmp,lswap)
  return tmp

n = 20
for j in range(10):
  example = [random.randrange(1,100) for i in range(n)]
  print "list:", example
  print "sorted:", bubble_sort_best(example)
  print

# Keep in mind that the optimization of remembering the last swap does
# not change the complexity in the worst case! Hence, it is just to
# speed things up a little, but it still remains quadratic.
