"""file: ex2.py -- correction td2 -- dichotomy searching and insertion
sort

Implements insertion sort with optimal insertion by dichotomy.

"""

import random

def dichotomy(li,start,end,x):
  """search for x in li between position start and end (included)
  
  li is supposed to be sorted

  """
  if end < start: return -1 # base case
  # else use dichotomy
  mid = start + (end - start)/2
  if li[mid] == x: return mid
  elif x < li[mid]: return dichotomy(li,start,mid-1,x)
  else: return dichotomy(li,mid+1,end,x)

def insert_sort_naive(li):
  tmp = li[:] # copy li
  # swap first two if necessary
  if tmp[0]>tmp[1]: tmp[0],tmp[1] = tmp[1],tmp[0]
  for i in range(2,len(tmp)):
    # invariant: tmp[:i] is supposed sorted
    # search for position of tmp[i] in tmp[:i] 
    j = i
    while tmp[j-1] > tmp[i]: j -= 1
    # shift pos by 1 of the elements tmp[j:i]
    x = tmp[i]
    while j <= i:
      x,tmp[j] = tmp[j],x
      j += 1
  return tmp

def insert_sort(li):
  """returns a sorted copy of li. 

  Uses insertion sort. Find position naively.

  """
  tmp = li[:] # copy li
  # swap first two if necessary
  if tmp[0]>tmp[1]: tmp[0],tmp[1] = tmp[1],tmp[0]
  for i in range(2,len(tmp)):
    # invariant: tmp[:i] is supposed sorted
    # search for position of tmp[i] in tmp[:i] by dichotomy
    j = i
    while j>=1 and tmp[j-1] > tmp[i]: j -= 1
    # shift pos by 1 of the elements tmp[j:i]
    x = tmp[i]
    while j <= i:
      x,tmp[j] = tmp[j],x
      j += 1
  return tmp

def dichotomy_mod(li,start,end,x):
  """li is supposed to be sorted. Its returns the position where x is
  supposed to be included at.

  """
  if end < start: return start # base case
  # else use dichotomy
  mid = start + (end - start)/2
  if li[mid] == x: return mid
  elif x < li[mid]: return dichotomy_mod(li,start,mid-1,x)
  else: return dichotomy_mod(li,mid+1,end,x)

def insert_sort_dicho(li):
  """returns a sorted copy of li. 

  Uses insertion sort. Uses dichotomy to find position.
  
  """
  tmp = li[:] # copy li
  # swap first two if necessary
  if tmp[0]>tmp[1]: tmp[0],tmp[1] = tmp[1],tmp[0]
  for i in range(2,len(tmp)):
    # invariant: tmp[:i] is supposed sorted
    # search for position of tmp[i] in tmp[:i] by dichotomy
    j = dichotomy_mod(tmp,0,i-1,tmp[i])
    # shift pos by 1 of the elements tmp[j:i]
    x = tmp[i]
    while j <= i:
      x,tmp[j] = tmp[j],x
      j += 1
  return tmp
  
# testing
n = 20
for j in range(10):
  example = [random.randrange(1,100) for i in range(n)]
  print "list:", example
  print "sorted:", insert_sort(example), "\n"
